import getpass
import random
import os
from pathlib import Path
from typing import List


def add_authorized_public_keys(public_keys: str = None):
    """
    Adds allowed public keys to ~/.ssh/authorized_keys
    Parameters:
        public_key(str): authorized public keys for ssh
    """
    
    if public_keys is None:
        public_keys = [str(getpass.getpass(prompt="authorized ssh-rsa pub auth keys sep by newline:"))]

    home = str(Path.home())
    ssh_path = home + "/.ssh/"
    if not os.path.exists(ssh_path):
        os.mkdir(ssh_path)
    try:
        with open(ssh_path + "authorized_keys", 'a+') as f:
            for public_key in public_keys:
                f.write(public_key)
                f.write('\n')
    except BaseException:
        raise RuntimeError("Error occured while adding keys")


def sshd_config(ssh_port=random.randint(49153, 65534),
                sshd_config_params: str = None) -> int:
    """
    Setup sshd_config as specified by config_params array
    Parameters:
        ssh_port(int): selects a random port from 49153 to 65534 unless specified
        sshd_config(str): custom sshd_config

    Returns:
        (int): ssh port number
    """

    if sshd_config_params is None:
        config_params = ["ClientAliveInterval 120",
                         "PasswordAuthentication no",
                         "PermitRootLogin yes",
                         "Port " + str(ssh_port)]
    else:
        config_params = [sshd_config_params]

    with open(r"/etc/ssh/sshd_config", 'w') as f:
        for config in config_params:
            f.write(config + '\n')
    return ssh_port


def ssh_config_params(hostname: str, user: str, ssh_port: int) -> List[str]:
    """
    Returns ssh client config for the given hostname, user and port
    """
    config_params = ["Host " + str(hostname),
                     "\tHostname %h",
                     "\tUser " + str(user),
                     "\tPort " + str(ssh_port),
                     "\tLogLevel ERROR",
                     "\tUserKnownHostsFile /dev/null",
                     "\tProxyCommand cloudflared access ssh --hostname %h"]
    return config_params


def ssh_config(config_params: str):
    """
    Configures ~/.ssh/config based on input config parameter
    Parameters
        config_params(str): Holds list of str paramters per line to be added to config
    """
    
    home = str(Path.home())
    ssh_path = home + "/.ssh/"
    with open(ssh_path + "config", "w+") as f:
        for config in config_params:
            f.write(config + "\n")



import random
import getpass
from CFsshTunnel.cloudflare.cloudflare_config import extract_tunnel_metrics
from CFsshTunnel.cloudflare.cloudflare import create_cloudflare_tunnel
from CFsshTunnel.utils.package_installer import apt_package_installer, deb_package_installer
from CFsshTunnel.ssh.ssh_config import add_authorized_public_keys, sshd_config, ssh_config_params
from CFsshTunnel.ssh.ssh import start_ssh_server
from CFsshTunnel.utils.decorated_print import box_border, seperator_command_border, seperator_config_border
from typing import List, Tuple

def CFsshTunnel(
    cloudflare_config_path: str = None,
    ssh_port=random.randint(
        49153,
        65534),
    sshd_config_params: str = None,
    public_keys: str = None,
    test: bool = False) -> Tuple[List[str], str, str]:
    """
    Configures and initiates server as specified by default/user
    Args
        cloudflare_config_path(str): path to custom cloudflare config.yaml
        ssh_port(int): specifies port that openssh-server is listening to
        sshd_config(str): specifies config for sshd_config
        public_keys(str): list of authorized public keys to be added to server ~/.ssh/authorized_keys
        keep_alive(bool): specifies where the server python program should keep running infdefinitely
    Returns
        ssh_configuration(List[str]): list of str of ssh config for client
        hostname(str): cloudflare tunnel hostname
        user(str): server username
    """
    # Check required packages on server and install if required
    apt_package_installer("openssh-server")

    # https://developers.cloudflare.com/cloudflare-one/connections/connect-apps/install-and-setup/installation
    cloudflare_deb_url = "https://github.com/cloudflare/cloudflared/releases/latest/download/cloudflared-linux-amd64.deb"
    deb_package_installer("cloudflared", cloudflare_deb_url)

    # accepts List[str] or just str
    add_authorized_public_keys(public_keys=public_keys)

    # update ssh_config as specified by user or default parameters and random
    # port if not
    ssh_port = sshd_config(
        ssh_port=ssh_port,
        sshd_config_params=sshd_config_params)

    # restarts openssh-server service with new ssh_config
    start_ssh_server()

    metrics_port = random.randint(49153, 65534)
    if metrics_port == ssh_port:
        metrics_port += 1

    ssh_cloudflare_call = "cloudflared tunnel --url ssh://localhost:" + str(ssh_port) +\
        " --logfile cloudflared.log --metrics localhost:" + \
        str(metrics_port)

    # create a trycloudflare.com free tunnel and route
    # ssh://localhost:ssh_port throught the assigned public domain
    create_cloudflare_tunnel(
        cloudflare_call=ssh_cloudflare_call,
        cloudflare_config_path=cloudflare_config_path)

    metrics_url = "http://localhost:" + str(metrics_port) + "/metrics"
    hostname = extract_tunnel_metrics(metrics_url)
    user = getpass.getuser()

    ssh_configuration = ssh_config_params(
        hostname=hostname, user=user, ssh_port=ssh_port)

    if test:
        ssh_configuration.insert(
            len(ssh_configuration),
            "\tStrictHostKeyChecking no")

    header = "open_ssh_cloudflare tunnel route is now alive at:\n" + hostname + "\n"
    box_border(header)
    print("Update ~/.ssh/config on client as below:\n")
    print("#Client ~/.ssh/config")
    seperator_config_border(ssh_configuration)
    print(
        "Note: Windows client users on PS/cmd, provide full path to cloudflared.exe in ProxyCommand\n\
        Also applies to linux users if PATH to cloudflared isn't added to $PATH\n\
        Ex: Instead of \n\
            `ProxyCommand cloudflared access ssh --hostname %h`\n\
        use `ProxyCommand <complete_path_to_cloudflare.exe> access ssh --hostname %h\n")
    print("\nConnect to openssh-server through the following public domain:")
    client_command = "$ ssh " + str(hostname)
    seperator_command_border(client_command)
    print("\nNote: Since user authentication through ssh-rsa key pair is configured to be true by default,\n\
        only those users whose public key has been added to the config will be able to access the server")
    return ssh_configuration, hostname, user


def keep_alive(state: bool = True):
    # keeps the server alive?
    while state:
        continue



import sys
import subprocess
import apt


def check_installed(package_name: str) -> bool:
    print("Checking for " + package_name)
    cache = apt.Cache()
    package_installed = False
    if package_name in cache:
        package_installed = cache[package_name].is_installed
    return package_installed


def apt_package_installer(package_name: str):
    """
    Checks for package and installs if needed
    Parameters
        package_name(str): name of the package to be installed
    """
    package_installed = check_installed(package_name=package_name)

    if package_installed:
        print("{pkg_name} already installed".format(pkg_name=package_name))
    else:
        print(
            "Installing {pkg_name} through apt-get".format(pkg_name=package_name))

        cache = apt.cache.Cache()
        cache.update()
        cache.open()
        package = cache[package_name]
        package.mark_install()
        try:
            cache.commit()
        except Exception as arg:
            print >> sys.stderr, "{pkg_name} installation failed [{err}]".format(
                pkg_name=package_name, err=str(arg))


def deb_package_installer(package_name: str, package_url: str):
    """
    Downloads and installs .deb pack from specified url
    """
    package_installed = check_installed(package_name=package_name)
    if package_installed:
        print("{pkg_name} already installed".format(pkg_name=package_name))
    else:
        print("Installing {pkg_name}".format(pkg_name=package_name))
        url_split = package_url.split('/')
        deb_name = url_split[-1]
        try:
            subprocess.call(["wget", package_url])
            subprocess.call(["sudo", "dpkg", "-i", deb_name])
            subprocess.call(["sudo", "rm", "-f", deb_name])
        except BaseException:
            raise RuntimeError(
                "Failed to install package {pkg_name}".format(
                    pkg_name=package_name))



import subprocess
import random
import time
from CFsshTunnel.utils.decorated_print import box_border, seperator_command_border
from CFsshTunnel.utils.package_installer import check_installed


def install_codeserver():
    """
        Installs code-server from https://code-server.dev/install.sh script
    """
    package_installed = check_installed("code-server")
    if package_installed is False:
        print("Installing code-server")
        curl_process = subprocess.Popen(
            "curl -fsSL https://code-server.dev/install.sh".split(" "),
            stdout=subprocess.PIPE)
        subprocess.check_output(("sh"), stdin=curl_process.stdout)
        curl_process.wait()


def launch_codeserver(user:str,
        hostname: str, 
        code_server_config_path: str = None,
        server_port: int = random.randint(
            49153,
            65534),
        client_port: int = 8080,
        ssh_mode: bool = True
):
    """
    Launch code-server on specified or random port and map it to specified cloudflare tunnel
    Args:
        user(str): server user
        hostname(str): cloudflare tunnel hostname
        code_server_config_path(str): path to custom config.yaml for code-server
        server_port(int): specify port for codeserver
        client_port(int): specify client port to connect to code-server
        ssh_mode(bool): mode of connection to code-server via ssh, if false
                        a new cloudflare tunnel is created and mapped to code-server,
                        passwd auth is enforced from default config and custom config.yaml is not allowed
                        to ensure safety
    """
    # install code-server if required
    install_codeserver()

    if ssh_mode is False:
        auth_mode = "password"
    else:
        auth_mode = "none"

    if code_server_config_path is None:
        #defaul code-server config file
        code_server_config_path = "~/.config/code-server/config.yaml"

    # code-server launch comand
    codeserver_command = "code-server" + " "\
         + "--bind-addr" + " " + "127.0.0.1:"+ str(server_port) + " "\
         + "--auth" + " " + auth_mode + " "\
         + "--config"+ " " + code_server_config_path
    
    # launch code-server
    print("Launching code-server...")
    subprocess.Popen(codeserver_command.split(" "))
    time.sleep(5)
    #port forward client localhost to remote instance
    print("Port forward client localhost:8080 onto the remote code-server instance through ssh as follows(run this on your client device): \n")
    seperator_command_border("ssh -N -L "+ str(client_port) + ":127.0.0.1:" + str(server_port) + " " + user + "@" + hostname + " " + "&")
    box_border("Codeserver is now accessible at: http://localhost:" + str(client_port) +"\n Note: localhost cannot provide SSL, so use http instead of https")


from typing import List


def box_border(text: str) -> None:
    """
        Adds a box border made of "=" around the passed text
    """
    print("\n")
    text_split = text.split(sep='\n')
    if len(text_split[len(text_split) - 1]) == 0:
        text_split.pop(len(text_split) - 1)
    max_text_length = len(str(max(text_split, key=len)))
    print("+" + "-" * (max_text_length + 2) + "+")
    for row in text_split:
        end_filler = " "
        if len(row) < max_text_length:
            end_filler *= max_text_length - len(row) + 1
        print("| " + row + end_filler + "|")
    print("+" + "-" * (max_text_length + 2) + "+")
    print("\n")


def seperator_config_border(config: List[str]) -> None:
    """
        Adds a seperator border made of "-" above and below the passed text, appended with "#" at the start
    """
    max_text_length = len(str(max(config, key=len)))
    print("#" + "-" * (max_text_length - 1))
    for row in config:
        print(row)
    print("#" + "-" * (max_text_length - 1))


def seperator_command_border(command: str) -> None:
    """
        Adds a seperator border made of "-" above and below the passed text
    """
    max_text_length = len(command)
    print("-" * (max_text_length))
    print(command)
    print("-" * (max_text_length))



import subprocess

def start_ssh_server():
    """
    Restart ssh
    """
    subprocess.run(["service", "ssh", "restart"])


def connect_to_server(hostname: str):
    """
    Checks connection to server hostname
    """
    status = subprocess.check_call(
        ["ssh", hostname, "ls"], shell=False, stdout=subprocess.PIPE)
    if status == 0:
        print("Connection Successful!")
    else:
        print("Failed to connect to remote server")



from CFsshTunnel.cloudflare.cloudflare_config import extract_tunnel_metrics
from CFsshTunnel.cloudflare.cloudflare import create_cloudflare_tunnel
from CFsshTunnel.utils.package_installer import apt_package_installer, deb_package_installer
from CFsshTunnel.ssh.ssh_config import ssh_config, add_authorized_public_keys, sshd_config
from CFsshTunnel.CFsshTunnel import CFsshTunnel
from CFsshTunnel.ssh.ssh import connect_to_server, start_ssh_server
from CFsshTunnel.utils.decorated_print import box_border, seperator_command_border, seperator_config_border
from CFsshTunnel.code_server.code_server import install_codeserver, launch_codeserver



import subprocess
import time

def create_cloudflare_tunnel(cloudflare_call: str = None,
                             cloudflare_config_path: str = None
                             ):
    """
    Creates cloudflare tunnel for specified localhost:ssh_port
    cloudflare logs stored under cloudflared.log,
    metrics at localhost:ssh_port+1

    Parameters
        cloudflare_call(str): cloudflared command to execute
        configured_cloudflare(bool): set to true if ~/.cloudflared/config.yaml
    """
    if cloudflare_config_path is None:
        cloudflare_config_path = "~/.cloudflared/config.yaml"

    configured_cloudflare_call = "cloudflared tunnel --config " + cloudflare_config_path 

    if cloudflare_call is None:
        cloudflare_call = configured_cloudflare_call

    cf_process = subprocess.Popen(
        cloudflare_call.split(" "),
        stdout=subprocess.PIPE,
        universal_newlines=True)
    time.sleep(15)

    if cf_process.poll() is not None:
        raise RuntimeError("Failed to create cloudflare tunnel\n")



import time
import urllib.request

def extract_tunnel_metrics(metrics_url: str) -> str:
    """
    Utility function to extract hostname from cloudflare tunnel metrics
    Parameters
        metrics_url(str): str of url to cloudflare metrics
    Returns
        hostname(str): returns tunnel hostname
    """
    hostname = None
    for _ in range(10):
        try:
            with urllib.request.urlopen(metrics_url) as response:
                response = str(response.read())
                hostname_begin_str = "userHostname=\"https://"
                hostname_end_str = "\"}"
                begin = response.find(hostname_begin_str)
                response = response[begin + len(hostname_begin_str):]
                end = response.find(hostname_end_str)
                hostname = response[:end]
        except BaseException:
            time.sleep(10)
        if hostname is None:
            raise RuntimeError(
                "Failed to get hostname from cloudflare metrics\n\n")
    return hostname



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='CFsshTunnel',
    packages=find_packages(),
    version='0.1.8',
    license='MIT',
    description='Cloudflare Tunnel for open-ssh and code-server',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/ThePilot916/CFsshTunnel.git',
    download_url='',
    keywords=[
        'SSH',
        'Tunnel',
        'Cloudflare',
        'colab_ssh',
        'aws',
        'aws_ssh',
        'code-server',
        'remote development'
        'vscode_server',
        'vscode_remote',
        'google colab'],
    install_requires=["python-apt"],
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8'])
